@@ -280,7 +280,7 @@ def function_measure_multi_dimensional_shifts(xdata: DataAndMetadata.DataAndMeta
             iteration_shape += (xdata.data_shape[i],)
             dimensional_calibrations.append(xdata.dimensional_calibrations[i])
         else:
-            intensity_calibration = xdata.dimensional_calibrations[i]
+            intensity_calibration = Calibration.Calibration(scale=xdata.dimensional_calibrations[i].scale, units=xdata.dimensional_calibrations[i].units)
 
     shape: typing.Tuple[int, ...]
     register_slice: typing.Union[slice, typing.Tuple[slice, slice]]
@@ -333,8 +333,21 @@ def function_measure_multi_dimensional_shifts(xdata: DataAndMetadata.DataAndMeta
         # Use a little bit more than half the CPU cores, but not more than 20 because then we actually get a slowdown
         # because of our HDF5 storage handler not being able to grant parallel access to the data
         num_threads = min(int(round(num_cpus * 0.6)), 20)
-    sections = list(range(start_index, navigation_len, max(1, navigation_len//num_threads)))
-    sections.append(navigation_len)
+
+    # Unfortunately multi-threading cannot be used when we cross-correlate with the first frame and max_shift
+    # is not None because in order to create the mask for each frame we need the shift from the previous frame.
+    # This does not work for the "block boundaries" where the data is split between the frames, so we have to
+    # do a single-threaded calculation in this case.
+    # If the shifts reference is not the first or last frame in the sequence, we can actually use two threads, both
+    # starting from reference_index and iterating away from it.
+    if max_shift is not None and reference_index is not None:
+        if reference_index == 0 or reference_index == navigation_len - 1:
+            sections = [start_index, navigation_len]
+        else:
+            sections = [start_index, reference_index, navigation_len]
+    else:
+        sections = list(range(start_index, navigation_len, max(1, navigation_len//num_threads)))
+        sections.append(navigation_len)
     barrier = threading.Barrier(len(sections))
 
     def run_on_thread(range_):
@@ -347,11 +360,11 @@ def function_measure_multi_dimensional_shifts(xdata: DataAndMetadata.DataAndMeta
                 coords = numpy.unravel_index(i, iteration_shape)
                 data_coords = coords[:shift_axis_indices[0]] + (...,) + coords[shift_axis_indices[0]:]
                 if reference_index is None:
-                    coords_ref = numpy.unravel_index(i - 1, iteration_shape)
+                    coords_ref = numpy.unravel_index(i - range_.step, iteration_shape)
                     data_coords_ref = coords_ref[:shift_axis_indices[0]] + (...,) + coords_ref[shift_axis_indices[0]:]
                     local_reference_data = xdata.data[data_coords_ref]
-                elif max_shift is not None and i > start_index:
-                    last_coords = numpy.unravel_index(i - 1, iteration_shape)
+                elif max_shift is not None and i != range_.start:
+                    last_coords = numpy.unravel_index(i - range_.step, iteration_shape)
                     last_shift = shifts[last_coords]
                     data_shape = local_reference_data[register_slice].shape
                     if len(data_shape) == 2:
@@ -362,8 +375,23 @@ def function_measure_multi_dimensional_shifts(xdata: DataAndMetadata.DataAndMeta
         finally:
             barrier.wait()
 
-    for i in range(len(sections) - 1):
-        threading.Thread(target=run_on_thread, args=(range(sections[i], sections[i+1]),)).start()
+    if max_shift is not None and reference_index is not None:
+        # Set up the threads for the case with max_shift and reference index: As explained above, we need a special
+        # setup because the result relies on the previous shift.
+        if len(sections) == 2:
+            if reference_index == 0:
+                threading.Thread(target=run_on_thread, args=(range(sections[0], sections[1]),)).start()
+            # Reference index is the last frame, so go backwards from there
+            else:
+                threading.Thread(target=run_on_thread, args=(range(sections[1] - 1, sections[0] - 1, -1),)).start()
+        else:
+            # If the reference index is somewhere inside the sequence, we can use two threads, one going from
+            # reference_index to 0 (backwards) and one gaing from reference_index to the end.
+            threading.Thread(target=run_on_thread, args=(range(sections[1], sections[0] - 1, -1),)).start()
+            threading.Thread(target=run_on_thread, args=(range(sections[1], sections[2]),)).start()
+    else:
+        for i in range(len(sections) - 1):
+            threading.Thread(target=run_on_thread, args=(range(sections[i], sections[i+1]),)).start()
     barrier.wait()
 
     # For debugging it is helpful to run a non-threaded version of the code. Comment out the 3 lines above and uncomment
@@ -1034,6 +1062,96 @@ class MakeTableauMenuItemDelegate:
         window.display_data_item(result_data_item)
 
 
+class AlignImageSequence(Symbolic.ComputationHandlerLike):
+    label = _("Align and integrate image sequence")
+    inputs = {"input_data_item": {"label": _("Input data item")},
+              "reference_index": {"label": _("Reference index for shifts")},
+              "relative_shifts": {"label": _("Measure shifts relative to previous slice")},
+              "max_shift": {"label": _("Max shift between consecutive frames (in pixels, <= 0 to disable)")},
+              "bounds_graphic": {"label": _("Shift bounds")},
+              }
+    outputs = {"shifts": {"label": _("Shifts")},
+               "integrated_sequence": {"label": _("Integrated sequence")},
+               }
+
+    def __init__(self, computation, **kwargs):
+        self.computation = computation
+
+    def execute(self, *, input_data_item: API.DataItem, reference_index: typing.Union[None, int, typing.Sequence[int]]=None, relative_shifts: bool=True, max_shift: int=0, bounds_graphic: typing.Optional[API.Graphic]=None):
+        input_xdata = input_data_item.xdata
+        bounds = None
+        if bounds_graphic is not None:
+            bounds = bounds_graphic.bounds
+        max_shift_ = max_shift if max_shift > 0 else None
+        reference_index = reference_index if not relative_shifts else None
+        shifts_xdata = function_measure_multi_dimensional_shifts(input_xdata, 'data', reference_index=reference_index, bounds=bounds, max_shift=max_shift_)
+        self.__shifts_xdata = Core.function_transpose_flip(shifts_xdata, transpose=True, flip_v=False, flip_h=False)
+        aligned_input_xdata = function_apply_multi_dimensional_shifts(input_xdata, shifts_xdata.data, 'data')
+        self.__integrated_input_xdata = Core.function_sum(aligned_input_xdata, axis=0)
+
+    def commit(self):
+        self.computation.set_referenced_xdata("shifts", self.__shifts_xdata)
+        self.computation.set_referenced_xdata("integrated_sequence", self.__integrated_input_xdata)
+
+
+class AlignImageSequenceMenuItemDelegate:
+
+    def __init__(self, api):
+        self.__api = api
+        self.menu_id = "processing_menu"  # required, specify menu_id where this item will go
+        self.menu_name = _("Processing")  # optional, specify default name if not a standard menu
+        self.menu_before_id = "window_menu"  # optional, specify before menu_id if not a standard menu
+        self.menu_item_name = _("[EXPERIMENTAL] Align image sequence")  # menu item name
+
+    def menu_item_execute(self, window: API.DocumentWindow):
+        try:
+            selected_data_item = window.target_data_item
+            error_msg = "Select one data item that contains a sequence or 1D-collection of two-dimensional data."
+            assert selected_data_item is not None, error_msg
+            assert selected_data_item.xdata is not None, error_msg
+            assert selected_data_item.xdata.is_sequence or selected_data_item.xdata.is_collection, error_msg
+            assert not (selected_data_item.xdata.is_sequence and selected_data_item.xdata.is_collection), error_msg
+            if selected_data_item.xdata.is_collection:
+                assert selected_data_item.xdata.collection_dimension_count == 1, error_msg
+            assert selected_data_item.xdata.datum_dimension_count == 2, error_msg
+
+            bounds_graphic = None
+            if selected_data_item.display.selected_graphics:
+                for graphic in selected_data_item.display.selected_graphics:
+                    if graphic.graphic_type in {"rect-graphic", "interval-graphic"}:
+                        bounds_graphic = graphic
+
+            # Make a result data item with 3 dimensions to ensure we get a large_format data item
+            result_data_item = self.__api.library.create_data_item_from_data(numpy.zeros((1,1,1)), title=f"{selected_data_item.title} aligned and integrated")
+            shifts = self.__api.library.create_data_item_from_data(numpy.zeros((2, 2)), title=f"{selected_data_item.title} measured shifts")
+
+            inputs = {"input_data_item": {"object": selected_data_item, "type": "data_source"},
+                      "reference_index": 0,
+                      "relative_shifts": False,
+                      "max_shift": 0,
+                      }
+            if bounds_graphic:
+                inputs["bounds_graphic"] = bounds_graphic
+
+            self.__api.library.create_computation("nion.align_and_integrate_image_sequence",
+                                                  inputs=inputs,
+                                                  outputs={"shifts": shifts,
+                                                           "integrated_sequence": result_data_item})
+            window.display_data_item(result_data_item)
+            window.display_data_item(shifts)
+
+            display_item = self.__api.library._document_model.get_display_item_for_data_item(shifts._data_item)
+            display_item.display_type = "line_plot"
+            display_item._set_display_layer_properties(0, stroke_color='#1E90FF', stroke_width=2, fill_color=None, label="y")
+            display_item._set_display_layer_properties(1, stroke_color='#F00', stroke_width=2, fill_color=None, label="x")
+
+        except Exception as e:
+            import traceback
+            traceback.print_exc()
+            from nion.swift.model import Notification
+            Notification.notify(Notification.Notification("nion.computation.error", "\N{WARNING SIGN} Computation", "Align sequence of images failed", str(e)))
+
+
 class MultiDimensionalProcessingExtension:
 
     extension_id = "nion.experimental.multi_dimensional_processing"
@@ -1045,6 +1163,7 @@ class MultiDimensionalProcessingExtension:
         self.__apply_shifts_menu_item_ref = api.create_menu_item(ApplyShiftsMenuItemDelegate(api))
         self.__crop_menu_item_ref = api.create_menu_item(CropMenuItemDelegate(api))
         self.__tableau_menu_item_ref = api.create_menu_item(MakeTableauMenuItemDelegate(api))
+        self.__align_image_sequence_menu_item_ref = api.create_menu_item(AlignImageSequenceMenuItemDelegate(api))
 
     def close(self):
         self.__integrate_menu_item_ref.close()
@@ -1057,6 +1176,8 @@ class MultiDimensionalProcessingExtension:
         self.__crop_menu_item_ref = None
         self.__tableau_menu_item_ref.close()
         self.__tableau_menu_item_ref = None
+        self.__align_image_sequence_menu_item_ref.close()
+        self.__align_image_sequence_menu_item_ref = None
 
 
 class AxisChoiceVariableHandler(Observable.Observable):
@@ -1234,3 +1355,4 @@ Symbolic.register_computation_type("nion.measure_shifts", MeasureShifts)
 Symbolic.register_computation_type("nion.apply_shifts", ApplyShifts)
 Symbolic.register_computation_type("nion.crop_multi_dimensional", Crop)
 Symbolic.register_computation_type("nion.make_tableau_image", MakeTableau)
+Symbolic.register_computation_type("nion.align_and_integrate_image_sequence", AlignImageSequence)
